<?php if (!defined('ABSPATH')) exit;
class FooEvents_Zoom_API_Helper {

    public  $Config;

    public function __construct($config) {

        $this->Config = $config;

        add_action('wp_ajax_fooevents_zoom_test_access', array($this, 'fooevents_zoom_test_access'));
        add_action('wp_ajax_fooevents_zoom_fetch_users', array($this, 'fooevents_zoom_fetch_users'));
        add_action('wp_ajax_fooevents_fetch_zoom_meeting', array($this, 'fooevents_fetch_zoom_meeting'));
        add_action('wp_ajax_fooevents_update_zoom_registration', array($this, 'fooevents_update_zoom_registration'));

    }

    /**
     * Generate a Zoom API JWT token based using the provided API key and secret
     * 
     * @param string $key
     * @param string $secret
     * @param int $expiry
     * 
     * @return string
     */
    private function fooevents_zoom_generate_jwt($key, $secret, $expiry = 300) {

        if ( trim($key) == '' || trim($secret) == '' ) {

            return '';

        }

        require_once($this->Config->vendorPath.'/php-jwt/BeforeValidException.php');
        require_once($this->Config->vendorPath.'/php-jwt/ExpiredException.php');
        require_once($this->Config->vendorPath.'/php-jwt/SignatureInvalidException.php');
        require_once($this->Config->vendorPath.'/php-jwt/JWT.php');

        $token = array(
            "iss" => $key,
            "exp" => time() + $expiry
        );

        $jwt = \Firebase\FooEvents_JWT\JWT::encode($token, $secret);

        return $jwt;

    }

    /**
     * Generate a Zoom API JWT token from the saved key and secret
     * 
     * @return string
     */
    private function fooevents_zoom_jwt() {

        $key = (string)get_option('globalWooCommerceEventsZoomAPIKey', '');
        $secret = (string)get_option('globalWooCommerceEventsZoomAPISecret', '');

        return $this->fooevents_zoom_generate_jwt($key, $secret);

    }

    /**
     * Perform a Zoom API request and return the result
     * 
     * @param string $method
     * @param array $args
     * @param string $jwt
     * @param string $request_type
     * 
     * @return array
     */
    private function fooevents_zoom_request($method = '', $args = array(), $jwt = '', $request_type = 'GET') {

        $result = array('status' => 'error');

        if ( $jwt == '' ) {

            $jwt = $this->fooevents_zoom_jwt();

        }

        if ( empty($jwt) ) {

            $result['message'] = __('Error generating JSON Web Token (JWT) using the provided Zoom API Key and API Secret.', 'woocommerce-events');

            return $result;

        }

        $params = http_build_query($args);

        $curl = curl_init();

        $curl_options = array(
            CURLOPT_URL => "https://api.zoom.us/v2/" . $method . ($request_type == 'GET' ? "?" . $params : ''),
            CURLOPT_RETURNTRANSFER => true,
            CURLOPT_ENCODING => "",
            CURLOPT_MAXREDIRS => 10,
            CURLOPT_TIMEOUT => 30,
            CURLOPT_HTTP_VERSION => CURL_HTTP_VERSION_1_1,
            CURLOPT_CUSTOMREQUEST => $request_type,
            CURLOPT_HTTPHEADER => array(
                "authorization: Bearer " . $jwt,
                "content-type: application/json"
            ),
        );

        if ( $request_type == 'POST' || $request_type == 'PUT' || $request_type == 'PATCH' ) {

            $curl_options[CURLOPT_POSTFIELDS] = json_encode($args);

        }

        curl_setopt_array($curl, $curl_options);

        $response = curl_exec($curl);
        $error = curl_error($curl);

        curl_close($curl);

        if ($error) {

            $result['message'] = __('Unable to connect to your Zoom account', 'woocommerce-events');

        } else {

            $result['status'] = 'success';

            if ( $response != "" ) {

                if ( $response_array = json_decode($response, true) ) {

                    if ( !empty($response_array['code']) && !empty($response_array['message']) ) {
    
                        $result = array(
                            'status' => 'error',
                            'message' => $response_array['message']
                        );
        
                    } else {
        
                        $result = array(
                            'status' => 'success',
                            'data' => $response_array
                        );
        
                    }
    
                }

            }

        }

        return $result;

    }

    /**
     * Tests whether or not the Zoom API key and secret have been set up correctly
     * 
     */
    public function fooevents_zoom_test_access() {
        
        $key = trim($_POST['key']);
        $secret = trim($_POST['secret']);

        if ( trim($key) == '' || trim($secret) == '' ) {

            $result = array(
                'status' => 'error',
                'message' => __('Please enter your Zoom API Key and API Secret', 'woocommerce-events')
            );

            echo json_encode($result);

            exit();

        }

        $jwt = $this->fooevents_zoom_generate_jwt($key, $secret);
        
        $result = $this->fooevents_zoom_request('users/me', array(), $jwt);

        if ( $result['status'] === 'success' ) {
            update_option('globalWooCommerceEventsZoomAdminID', $result['data']['id']);
        }

        echo json_encode($result);

        exit();

    }

    /**
     * Fetches all users on the account to allow selecting of specific users' meetings to display
     */
    public function fooevents_zoom_fetch_users() {

        $key = '';
        $secret = '';

        if ( !empty($_POST['key']) && trim($_POST['key']) != '' && !empty($_POST['secret']) && trim($_POST['secret']) != '' ) {
            
            $key = trim($_POST['key']);
            $secret = trim($_POST['secret']);

        }

        if ( trim($key) == '' || trim($secret) == '' ) {

            // Check for saved key and secret
            $key = get_option('globalWooCommerceEventsZoomAPIKey', '');
            $secret = get_option('globalWooCommerceEventsZoomAPISecret', '');

        }
        
        if ( trim($key) == '' || trim($secret) == '' ) {

            $result = array(
                'status' => 'error',
                'message' => __('Please enter your Zoom API Key and API Secret', 'woocommerce-events')
            );

            echo json_encode($result);

            exit();

        }

        $jwt = $this->fooevents_zoom_generate_jwt($key, $secret);

        $zoomUsers = array();
        $loadedAllUsers = false;
        $page = 1;
        $userCount = 0;

        while ( !$loadedAllUsers ) {

            $response = $this->fooevents_zoom_request('users', array(
                'page_size' => 300,
                'page_number' => $page,
                'status' => 'active'
            ), $jwt);

            if ( $response['status'] == 'success' ) {

                $users = array();

                foreach ( $response['data']['users'] as $user ) {

                    $userCount++;

                    if ( $user['type'] == 1 ) {
                        continue;
                    }

                    $users[$user['id']] = array(
                        'id' => $user['id'],
                        'first_name' => ucwords($user['first_name']),
                        'last_name' => ucwords($user['last_name']),
                        'email' => $user['email']
                    );

                }
                
                if ( empty($zoomUsers) ) {

                    $zoomUsers = $response;

                    $zoomUsers['data']['users'] = $users;

                } else {

                    $zoomUsers['data']['users'] = array_merge($zoomUsers['data']['users'], $users);

                }

                if ( $zoomUsers['data']['total_records'] > $userCount ) {

                    $page++;

                } else {

                    $loadedAllUsers = true;

                }

            } else {

                echo json_encode($response);

                exit();

            }

        }

        uasort($zoomUsers['data']['users'], function($a, $b) {
            return $a['first_name'] <=> $b['first_name'];
        });

        echo json_encode($zoomUsers);

        exit();

    }

    /**
     * Fetch available Zoom meetings
     * 
     * @return array
     */
    public function fooevents_fetch_zoom_meetings() {

        return $this->fooevents_fetch_zoom('meetings');

    }

    /**
     * Fetch available Zoom webinars
     * 
     * @return array
     */
    public function fooevents_fetch_zoom_webinars() {

        return $this->fooevents_fetch_zoom('webinars');

    }
    
    /**
     * Fetch available Zoom meetings/webinars
     * 
     * @return array
     */
    public function fooevents_fetch_zoom($endpoint = 'webinars') {

        $zoomMeetings = array();

        $date_format = get_option('date_format');
        $time_format = get_option('time_format');
        $globalWooCommerceEventsZoomUsers = json_decode(get_option('globalWooCommerceEventsZoomUsers', json_encode(array())), true);
        $globalWooCommerceEventsZoomSelectedUserOption = get_option('globalWooCommerceEventsZoomSelectedUserOption');
        $globalWooCommerceEventsZoomSelectedUsers = get_option('globalWooCommerceEventsZoomSelectedUsers');

        if ( empty($globalWooCommerceEventsZoomSelectedUserOption) || (!empty($globalWooCommerceEventsZoomSelectedUserOption) && $globalWooCommerceEventsZoomSelectedUserOption == 'me') || (!empty($globalWooCommerceEventsZoomSelectedUserOption) && $globalWooCommerceEventsZoomSelectedUserOption == 'select' && empty($globalWooCommerceEventsZoomSelectedUsers)) ) {

            $globalWooCommerceEventsZoomSelectedUsers = array('me');

        }

        foreach ( $globalWooCommerceEventsZoomSelectedUsers as $userID ) {

            $user = array(
                'id' => 'me',
                'first_name' => '',
                'last_name' => '',
                'email' => 'me'
            );

            if ( $userID != 'me' ) {

                $user = $globalWooCommerceEventsZoomUsers[$userID];

            }

            $loadedAllMeetings = false;
            $page = 1;

            while ( !$loadedAllMeetings ) {
                $response = $this->fooevents_zoom_request('users/' . $user['email'] . '/' . $endpoint, array(
                    'page_size' => 300,
                    'page_number' => $page
                ));
    
                if ( $response['status'] == 'success' ) {

                    $meetings = array();

                    foreach ( $response['data'][$endpoint] as &$zoomMeeting ) {

                        if ( $zoomMeeting['type'] == 1 ) {
                            continue;
                        }
                        
                        $zoomMeeting['id'] = $zoomMeeting['id'] . '_' . $endpoint;
    
                        if ( $zoomMeeting['type'] != 3 && $zoomMeeting['type'] != 6 ) {

                            $start_timestamp = strtotime($zoomMeeting['start_time']);
        
                            $start_date = new DateTime("@".$start_timestamp);
                            $start_date_timezone = new DateTimeZone($zoomMeeting['timezone']);
                            $start_date->setTimezone($start_date_timezone);
            
                            $zoomMeeting['start_date_display'] = $start_date->format($date_format);
                            $zoomMeeting['start_time_display'] = $start_date->format($time_format . ' T');

                        }

                        $zoomMeeting['host'] = $user;

                        $meetings[] = $zoomMeeting;
    
                    }
    
                    if ( empty($zoomMeetings) ) {
                        $zoomMeetings = $response;

                        $zoomMeetings['data'][$endpoint] = $meetings;
                    } else {
                        $zoomMeetings['data'][$endpoint] = array_merge($zoomMeetings['data'][$endpoint], $meetings);
                    }
    
                    if ( $zoomMeetings['data']['total_records'] > count($zoomMeetings['data'][$endpoint]) ) {
                        $page++;
                    } else {
                        $loadedAllMeetings = true;
                    }
                } else {
                    break;
                }
            }

        }

        if ( empty($zoomMeetings) ) {
            $zoomMeetings['status'] = 'error';
        }

        $zoomMeetings['user_count'] = count($globalWooCommerceEventsZoomSelectedUsers);

        return $zoomMeetings;

    }

    /**
     * Fetch individual Zoom meeting AJAX call
     * 
     */
    public function fooevents_fetch_zoom_meeting() {

        $zoomMeetingID = $_POST['zoomMeetingID'];

        $result = $this->do_fooevents_fetch_zoom_meeting($zoomMeetingID);

        echo json_encode($result);

        exit();

    }

    /**
     * Fetch individual Zoom meeting
     * 
     */
    public function do_fooevents_fetch_zoom_meeting($zoomMeetingID) {

        $zoomIDParts = explode("_", $zoomMeetingID);
        $zoomID = $zoomIDParts[0];
        $endpoint = !empty($zoomIDParts[1]) ? $zoomIDParts[1] : 'webinars';

        $result = $this->fooevents_zoom_request($endpoint . '/' . $zoomID);
        
        if ( $result['status'] == 'success' && !empty($result['data']) ) {

            $zoomMeeting = &$result['data'];

            $start_timestamp = 0;

            if ( $zoomMeeting['type'] != 5 && $zoomMeeting['type'] != 2 ) {

                // Recurrence type
                switch ( $zoomMeeting['recurrence']['type'] ) {
                    case 1:
                        
                        $zoomMeeting['recurrence']['type_display'] = $zoomMeeting['recurrence']['repeat_interval'] == 1 ? __('Daily', 'woocommerce-events') : __('Every', 'woocommerce-events') . ' ' . $zoomMeeting['recurrence']['repeat_interval'] . ' ' . __('days', 'woocommerce-events');
                        break;

                    case 2:
                        
                        $zoomMeeting['recurrence']['type_display'] = $zoomMeeting['recurrence']['repeat_interval'] == 1 ? __('Weekly', 'woocommerce-events') : __('Every', 'woocommerce-events') . ' ' . $zoomMeeting['recurrence']['repeat_interval'] . ' ' . __('weeks', 'woocommerce-events');
                        break;

                    case 3:

                        $zoomMeeting['recurrence']['type_display'] = $zoomMeeting['recurrence']['repeat_interval'] == 1 ? __('Monthly', 'woocommerce-events') : __('Every', 'woocommerce-events') . ' ' . $zoomMeeting['recurrence']['repeat_interval'] . ' ' . __('months', 'woocommerce-events');
                        break;

                }

                // Weekly days
                if ( isset($zoomMeeting['recurrence']['weekly_days']) ) {

                    $weekly_days = explode(',', $zoomMeeting['recurrence']['weekly_days']);

                    $zoomMeeting['recurrence']['type_display'] .= __(' on ', 'woocommerce-events');

                    $last_weekly_day = end($weekly_days);

                    foreach ( $weekly_days as $weekly_day ) {
                        
                        switch ( $weekly_day ) {
                            case 1:

                                $zoomMeeting['recurrence']['type_display'] .= __('Sunday', 'woocommerce-events');
                                break;

                            case 2:
                                
                                $zoomMeeting['recurrence']['type_display'] .= __('Monday', 'woocommerce-events');
                                break;

                            case 3:
                                
                                $zoomMeeting['recurrence']['type_display'] .= __('Tuesday', 'woocommerce-events');
                                break;

                            case 4:
                                
                                    $zoomMeeting['recurrence']['type_display'] .= __('Wednesday', 'woocommerce-events');
                                    break;

                            case 5:
                                    
                                    $zoomMeeting['recurrence']['type_display'] .= __('Thursday', 'woocommerce-events');
                                    break;

                            case 6:

                                $zoomMeeting['recurrence']['type_display'] .= __('Friday', 'woocommerce-events');
                                break;

                            case 7:
                                
                                $zoomMeeting['recurrence']['type_display'] .= __('Saturday', 'woocommerce-events');
                                break;

                        }

                        if ( $weekly_day != $last_weekly_day ) {

                            $zoomMeeting['recurrence']['type_display'] .= ', ';

                        }

                    }

                }

                // Monthly day
                if ( isset($zoomMeeting['recurrence']['monthly_day']) ) {

                    $zoomMeeting['recurrence']['type_display'] .= __(' on the ', 'woocommerce-events');
                    $zoomMeeting['recurrence']['type_display'] .= $zoomMeeting['recurrence']['monthly_day'];
                    $zoomMeeting['recurrence']['type_display'] .= __(' of the month', 'woocommerce-events');

                }

                // Monthly week
                if ( isset($zoomMeeting['recurrence']['monthly_week']) ) {

                    switch ( $zoomMeeting['recurrence']['monthly_week'] ) {
                        case -1:
                            
                            $zoomMeeting['recurrence']['type_display'] .= ' ' . __('on the last', 'woocommerce-events');
                            break;

                        case 1:
                            
                            $zoomMeeting['recurrence']['type_display'] .= ' ' . __('on the first', 'woocommerce-events');
                            break;

                        case 2:
                            
                            $zoomMeeting['recurrence']['type_display'] .= ' ' . __('on the second', 'woocommerce-events');
                            break;

                        case 3:
                            
                            $zoomMeeting['recurrence']['type_display'] .= ' ' . __('on the third', 'woocommerce-events');
                            break;

                        case 4:
                            
                            $zoomMeeting['recurrence']['type_display'] .= ' ' . __('on the fourth', 'woocommerce-events');
                            break;

                    }

                    // Monthly week day
                    if ( isset($zoomMeeting['recurrence']['monthly_week_day']) ) {

                        switch ( $zoomMeeting['recurrence']['monthly_week_day'] ) {
                            case 1:
                                
                                $zoomMeeting['recurrence']['type_display'] .= ' ' . __('Sunday', 'woocommerce-events');
                                break;

                            case 2:
                                
                                $zoomMeeting['recurrence']['type_display'] .= ' ' . __('Monday', 'woocommerce-events');
                                break;

                            case 3:
                                
                                $zoomMeeting['recurrence']['type_display'] .= ' ' . __('Tuesday', 'woocommerce-events');
                                break;

                            case 4:
                                
                                $zoomMeeting['recurrence']['type_display'] .= ' ' . __('Wednesday', 'woocommerce-events');
                                break;

                            case 5:
                                
                                $zoomMeeting['recurrence']['type_display'] .= ' ' . __('Thursday', 'woocommerce-events');
                                break;

                            case 6:
                                
                                $zoomMeeting['recurrence']['type_display'] .= ' ' . __('Friday', 'woocommerce-events');
                                break;

                            case 7:
                                
                                $zoomMeeting['recurrence']['type_display'] .= ' ' . __('Saturday', 'woocommerce-events');
                                break;

                        }

                    }

                }

                if ( !empty($zoomMeeting['recurrence']['end_date_time']) ) {

                    $end_timestamp = strtotime($zoomMeeting['recurrence']['end_date_time']);

                    $end_date = new DateTime("@".$end_timestamp);
                    $end_date_timezone = new DateTimeZone($zoomMeeting['timezone']);
                    $end_date->setTimezone($end_date_timezone);

                    $zoomMeeting['recurrence']['type_display'] .= ' ' . __('until', 'woocommerce-events') . ' ' . $end_date->format(get_option('date_format'));

                } elseif ( !empty($zoomMeeting['recurrence']['end_times']) ) {

                    $occurrences = (int)$zoomMeeting['recurrence']['end_times'];

                    $zoomMeeting['recurrence']['type_display'] .= ', ' . $occurrences . ' ' . ($occurrences == 1 ? __('occurrence', 'woocommerce-events') : __('occurrences', 'woocommerce-events'));

                }

                if ( !empty($zoomMeeting['occurrences']) ) {

                    foreach ( $zoomMeeting['occurrences'] as &$occurrence ) {

                        $start_timestamp = strtotime($occurrence['start_time']);
                        $start_date = new DateTime("@".$start_timestamp);
                        $start_date_timezone = new DateTimeZone($zoomMeeting['timezone']);
                        $start_date->setTimezone($start_date_timezone);

                        $occurrence['start_date_display'] = $start_date->format(get_option('date_format'));
                        $occurrence['start_time_display'] = $start_date->format(get_option('time_format') . ' T');

                        $end_timestamp = $start_timestamp + ((int)$occurrence['duration'] * 60);
                        $end_date = new DateTime("@".$end_timestamp);
                        $end_date_timezone = new DateTimeZone($zoomMeeting['timezone']);
                        $end_date->setTimezone($end_date_timezone);

                        $occurrence['end_time_display'] = $end_date->format(get_option('time_format') . ' T');

                        $occurrence['duration_display'] = $this->fooevents_format_minutes((int)$occurrence['duration']);

                    }

                    $zoomMeeting['start_date_display'] = $zoomMeeting['occurrences'][0]['start_date_display'];
                    $zoomMeeting['start_time_display'] = $zoomMeeting['occurrences'][0]['start_time_display'];
                    $zoomMeeting['end_time_display'] = $zoomMeeting['occurrences'][0]['end_time_display'];
                    $zoomMeeting['duration_display'] = $zoomMeeting['occurrences'][0]['duration_display'];

                }

            } else {

                if ( $zoomMeeting['type'] != 3 && $zoomMeeting['type'] != 6 ) {

                    $start_timestamp = strtotime($zoomMeeting['start_time']);
                    $start_date = new DateTime("@".$start_timestamp);
                    $start_date_timezone = new DateTimeZone($zoomMeeting['timezone']);
                    $start_date->setTimezone($start_date_timezone);

                    $zoomMeeting['start_date_display'] = $start_date->format(get_option('date_format'));
                    $zoomMeeting['start_time_display'] = $start_date->format(get_option('time_format') . ' T');

                    $end_timestamp = $start_timestamp + ((int)$zoomMeeting['duration'] * 60);
                    $end_date = new DateTime("@".$end_timestamp);
                    $end_date_timezone = new DateTimeZone($zoomMeeting['timezone']);
                    $end_date->setTimezone($end_date_timezone);

                    $zoomMeeting['end_time_display'] = $end_date->format(get_option('time_format') . ' T');

                    $zoomMeeting['duration_display'] = $this->fooevents_format_minutes((int)$zoomMeeting['duration']);

                }

            }

            $zoomMeeting['meeting_capacity'] = (!empty($zoomMeeting['settings']['registrants_restrict_number']) && (int)$zoomMeeting['settings']['registrants_restrict_number'] > 0) ? $zoomMeeting['settings']['registrants_restrict_number'] : $this->fooevents_zoom_user_meeting_capacity($zoomMeeting['host_id'], in_array((int)$zoomMeeting['type'], array(2, 3, 8)));
            
            $registrants = $this->fooevents_get_zoom_meeting_registrants($zoomMeetingID);

            $zoomMeeting['registrants'] = array(
                'total_records' => 0,
                'registrants' => array()
            );

            if ( $registrants['status'] == 'success' ) {

                $zoomMeeting['registrants'] = $registrants['data'];

            }
            
        }

        return $result;

    }

    /**
     * Format number of minutes into a presentable string of hours and minutes
     * 
     * @param int $minutes
     * 
     * @return string
     */
    private function fooevents_format_minutes($minutes) {

        $formatted_minutes = '';

        if ( $minutes >= 60 ) {

            $hours = floor($minutes / 60);
            $remaining_minutes = $minutes % 60;

            $formatted_minutes = $hours . ' ' . ($hours == 1 ? __('hour', 'woocommerce-events') : __('hours', 'woocommerce-events')) . ($remaining_minutes > 0 ? ' ' . $remaining_minutes . ' ' . __('minutes', 'woocommerce-events') : '');

        } else {

            $formatted_minutes = $minutes . ' ' . __('minutes', 'woocommerce-events');

        }

        return $formatted_minutes;

    }

    /**
     * Register attendee for a Zoom meeting
     * 
     * @param int $zoomMeetingID
     * @param array $args
     * 
     * @return array
     */
    public function fooevents_register_zoom_attendee($zoomMeetingID, $args) {

        $zoomIDParts = explode("_", $zoomMeetingID);
        $zoomID = $zoomIDParts[0];
        $endpoint = !empty($zoomIDParts[1]) ? $zoomIDParts[1] : 'webinars';

        $result = $this->fooevents_zoom_request($endpoint . '/' . $zoomID . '/registrants', $args, '', 'POST');

        return $result;
    }

    /**
     * Updates attendee registration statuses for a Zoom meeting
     * 
     * @param int $zoomMeetingID
     * @param array $args
     * 
     * @return array
     */
    public function fooevents_update_zoom_registration_statuses($zoomMeetingID, $args) {

        $zoomIDParts = explode("_", $zoomMeetingID);
        $zoomID = $zoomIDParts[0];
        $endpoint = !empty($zoomIDParts[1]) ? $zoomIDParts[1] : 'webinars';

        $result = $this->fooevents_zoom_request($endpoint . '/' . $zoomID . '/registrants/status', $args, '', 'PUT');

        return $result;

    }

    /**
     * Update Zoom registration and approval type
     * 
     */
    public function fooevents_update_zoom_registration() {

        $zoomMeetingID = $_POST['zoomMeetingID'];
        $recurringMeeting = (bool)$_POST['recurringMeeting'];

        $args = array(
            'settings' => array(
                'approval_type' => 0
            )
        );

        if ( $recurringMeeting ) {

            $args['settings']['registration_type'] = 1;

        }

        $zoomIDParts = explode("_", $zoomMeetingID);
        $zoomID = $zoomIDParts[0];
        $endpoint = !empty($zoomIDParts[1]) ? $zoomIDParts[1] : 'webinars';

        $result = $this->fooevents_zoom_request($endpoint . '/' . $zoomID, $args, '', 'PATCH');

        echo json_encode($result);

        exit();

    }

    /**
     * Fetch the Zoom user's maximum meeting capacity
     * 
     * @return int
     */
    public function fooevents_zoom_user_meeting_capacity($host_id = 'me', $is_meeting = true) {

        $capacity = 0;
        $result = $this->fooevents_zoom_request('users/' . $host_id . '/settings');

        if ( $result['status'] === 'success' ) {

            $capacity_key = $is_meeting ? 'meeting_capacity' : 'webinar_capacity';
            $capacity = $result['data']['feature'][$capacity_key];

        }

        return $capacity;

    }

    /**
     * Get the Zoom meeting registrants
     * 
     * @param int $zoomMeetingID
     * 
     * @return array
     */
    private function fooevents_get_zoom_meeting_registrants($zoomMeetingID) {

        $zoomIDParts = explode("_", $zoomMeetingID);
        $zoomID = $zoomIDParts[0];
        $endpoint = !empty($zoomIDParts[1]) ? $zoomIDParts[1] : 'webinars';

        $loadedAllRegistrants = false;
        $page = 1;

        $meetingRegistrants = array();

        while ( !$loadedAllRegistrants ) {
            
            $response = $this->fooevents_zoom_request($endpoint . '/' . $zoomID . '/registrants', array(
                'page_size' => 300,
                'page_number' => $page
            ));

            if ( $response['status'] == 'success' ) {

                if ( empty($meetingRegistrants) ) {

                    $meetingRegistrants = $response;

                } else {

                    $meetingRegistrants['data']['registrants'] = array_merge($meetingRegistrants['data']['registrants'], $response['data']['registrants']);

                }

                if ( $meetingRegistrants['data']['total_records'] > count($meetingRegistrants['data']['registrants']) ) {

                    $page++;

                } else {

                    $loadedAllRegistrants = true;

                }

            } else {

                return $response;

            }
        }

        return $meetingRegistrants;

    }

    /**
     * Adds attendees as registrants or updates registrant statuses for the Zoom meeting
     * 
     * @param int $order_id
     */
    public function add_update_zoom_registrants($order_id) {
        
        $tickets_query = new WP_Query( array('post_type' => array('event_magic_tickets'), 'posts_per_page' => -1, 'meta_query' => array( array( 'key' => 'WooCommerceEventsOrderID', 'value' => $order_id ) )) );
        $orderTickets = $tickets_query->get_posts();

        foreach ( $orderTickets as $ticket ) {

            $ticket_meta = get_post_meta($ticket->ID);

            $args = array(
                'email' => $ticket_meta['WooCommerceEventsAttendeeEmail'][0],
                'first_name' => $ticket_meta['WooCommerceEventsAttendeeName'][0],
                'last_name' => $ticket_meta['WooCommerceEventsAttendeeLastName'][0],
                'phone' => !empty($ticket_meta['WooCommerceEventsAttendeeTelephone']) ? $ticket_meta['WooCommerceEventsAttendeeTelephone'][0] : '',
                'org' => !empty($ticket_meta['WooCommerceEventsAttendeeCompany']) ? $ticket_meta['WooCommerceEventsAttendeeCompany'][0] : '',
                'job_title' => !empty($ticket_meta['WooCommerceEventsAttendeeDesignation']) ? $ticket_meta['WooCommerceEventsAttendeeDesignation'][0] : ''
            );

            $WooCommerceEventsProductID = $ticket_meta['WooCommerceEventsProductID'][0];
            $WooCommerceEventsType = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsType', true);
            $WooCommerceEventsZoomMultiOption = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomMultiOption', true);
            
            if ( empty($WooCommerceEventsZoomMultiOption) || $WooCommerceEventsZoomMultiOption == 'single' ) {

                // Single meeting
                $zoomMeetingID = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomWebinar', true);

                $this->add_update_single_zoom_registrant($zoomMeetingID, $args);

            } elseif ( $WooCommerceEventsZoomMultiOption == 'multi' ) {

                // Multiple meetings
                $WooCommerceEventsZoomWebinarMulti = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomWebinarMulti', true);

                foreach ( $WooCommerceEventsZoomWebinarMulti as $zoomMeetingID ) {

                    $this->add_update_single_zoom_registrant($zoomMeetingID, $args);

                }

            } elseif ( $WooCommerceEventsZoomMultiOption == 'bookings' ) {

                // Bookable slot meeting
                $WooCommerceEventsBookingOptionsJSON = get_post_meta($WooCommerceEventsProductID, 'fooevents_bookings_options_serialized', true);
                $WooCommerceEventsBookingOptions = json_decode(!empty($WooCommerceEventsBookingOptionsJSON) ? $WooCommerceEventsBookingOptionsJSON : '{}', true);
                $WooCommerceEventsBookingSlotID = $ticket_meta['WooCommerceEventsBookingSlotID'][0];
                $WooCommerceEventsBookingDateID = $ticket_meta['WooCommerceEventsBookingDateID'][0];

                if ( isset($WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['zoom_id']) ) {
    
                    $zoomMeetingID = $WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['add_date'][$WooCommerceEventsBookingDateID]['zoom_id'];

                    $this->add_update_single_zoom_registrant($zoomMeetingID, $args);

                }
                
            }

        }

    }

    /**
     * Adds a single attendee as registrant or updates registrant status for the provided Zoom meeting
     * 
     * @param int $zoomMeetingID
     * @param array $args
     */
    public function add_update_single_zoom_registrant($zoomMeetingID, $args) {

        if ( $zoomMeetingID != '' ) {

            if ( $args['email'] != '' && $args['first_name'] != '' && $args['last_name'] != '' ) {

                $result = $this->fooevents_register_zoom_attendee($zoomMeetingID, $args);

                if ( $result['status'] == 'error' ) {

                    // Possibly already exists, try updating to approved
                    $update_args = array(
                        'action' => 'approve',
                        'registrants' => array(
                            array('email' => $args['email'])
                        )
                    );

                    $result = $this->fooevents_update_zoom_registration_statuses($zoomMeetingID, $update_args);

                }

            }

        }

    }

    /**
     * Cancel registrations for all provided tickets
     * 
     * @param array $tickets
     */
    public function cancel_zoom_registrations($tickets) {

        $zoomRegistrants = array();

        foreach ( $tickets as $ticket ) {

            $WooCommerceEventsProductID = get_post_meta($ticket->ID, 'WooCommerceEventsProductID', true);
            $WooCommerceEventsType = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsType', true);
            $WooCommerceEventsZoomMultiOption = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomMultiOption', true);
            $WooCommerceEventsStatus = get_post_meta($ticket->ID, 'WooCommerceEventsStatus', true);

            $email = get_post_meta($ticket->ID, 'WooCommerceEventsAttendeeEmail', true);

            if ( $email == '' ) {

                $email = get_post_meta($ticket->ID, 'WooCommerceEventsPurchaserEmail', true);

            }

            if ( empty($WooCommerceEventsZoomMultiOption) || $WooCommerceEventsZoomMultiOption == 'single' ) {

                // Single meeting
                $zoomMeetingID = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomWebinar', true);
                
                if ( $zoomMeetingID != '' && !empty($WooCommerceEventsStatus) && $WooCommerceEventsStatus == 'Canceled' ) {

                    if ( empty($zoomRegistrants[(string)$zoomMeetingID]) ) {

                        $zoomRegistrants[(string)$zoomMeetingID] = array();

                    }
    
                    $zoomRegistrants[(string)$zoomMeetingID][] = array('email' => $email);

                }

            } elseif ( $WooCommerceEventsZoomMultiOption == 'multi' ) {

                // Multiple meetings
                $WooCommerceEventsZoomWebinarMulti = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomWebinarMulti', true);
                $WooCommerceEventsMultidayStatus = json_decode(get_post_meta($ticket->ID, "WooCommerceEventsMultidayStatus", true), true);

                for ( $i = 1; $i <= count($WooCommerceEventsZoomWebinarMulti); $i++ ) {

                    $zoomMeetingID = $WooCommerceEventsZoomWebinarMulti[$i - 1];

                    if ( $zoomMeetingID != '' && ($WooCommerceEventsStatus == 'Canceled' || empty($WooCommerceEventsMultidayStatus) || (!empty($WooCommerceEventsMultidayStatus) && $WooCommerceEventsMultidayStatus[(string)$i] == 'Canceled')) ) {

                        if ( empty($zoomRegistrants[(string)$zoomMeetingID]) ) {

                            $zoomRegistrants[(string)$zoomMeetingID] = array();

                        }
        
                        $zoomRegistrants[(string)$zoomMeetingID][] = array('email' => $email);

                    }

                }

            } elseif ( $WooCommerceEventsZoomMultiOption == 'bookings' ) {

                // Bookable slot meeting
                $WooCommerceEventsBookingOptionsJSON = get_post_meta($WooCommerceEventsProductID, 'fooevents_bookings_options_serialized', true);
                $WooCommerceEventsBookingOptions = json_decode(!empty($WooCommerceEventsBookingOptionsJSON) ? $WooCommerceEventsBookingOptionsJSON : '{}', true);
                $WooCommerceEventsBookingSlotID = $ticket_meta['WooCommerceEventsBookingSlotID'][0];
                $WooCommerceEventsBookingDateID = $ticket_meta['WooCommerceEventsBookingDateID'][0];

                if ( isset($WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['zoom_id']) ) {

                    $zoomMeetingID = $WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['add_date'][$WooCommerceEventsBookingDateID]['zoom_id'];

                    $zoomRegistrants[(string)$zoomMeetingID][] = array('email' => $email);

                }

            }

        }

        if ( !empty($zoomRegistrants) ) {

            foreach ( $zoomRegistrants as $zoomMeetingID => $registrants ) {

                $args = array(
                    'action' => 'cancel',
                    'registrants' => $registrants
                );
    
                $result = $this->fooevents_update_zoom_registration_statuses($zoomMeetingID, $args);

            }

        }

    }

    /**
     * Register an attendee for a meeting when manually saving a ticket
     * 
     * @param int $ticket_ID
     */
    public function register_ticket_attendee($ticket_ID) {

        $ticket_meta = get_post_meta($ticket_ID);

        $args = array(
            'email' => !empty($ticket_meta['WooCommerceEventsAttendeeEmail'][0]) ? $ticket_meta['WooCommerceEventsAttendeeEmail'][0] : $ticket_meta['WooCommerceEventsPurchaserEmail'][0],
            'first_name' => !empty($ticket_meta['WooCommerceEventsAttendeeName'][0]) ? $ticket_meta['WooCommerceEventsAttendeeName'][0] : $ticket_meta['WooCommerceEventsPurchaserFirstName'][0],
            'last_name' => !empty($ticket_meta['WooCommerceEventsAttendeeLastName'][0]) ? $ticket_meta['WooCommerceEventsAttendeeLastName'][0] : $ticket_meta['WooCommerceEventsPurchaserLastName'][0]
        );

        $WooCommerceEventsType = get_post_meta($ticket_meta['WooCommerceEventsProductID'][0], 'WooCommerceEventsType', true);
        $WooCommerceEventsZoomMultiOption = get_post_meta($ticket_meta['WooCommerceEventsProductID'][0], 'WooCommerceEventsZoomMultiOption', true);

        if ( empty($WooCommerceEventsZoomMultiOption) || $WooCommerceEventsZoomMultiOption == 'single' ) {

            // Single meeting
            if ( $ticket_meta['WooCommerceEventsStatus'][0] != 'Canceled' ) {

                $zoomMeetingID = get_post_meta($ticket_meta['WooCommerceEventsProductID'][0], 'WooCommerceEventsZoomWebinar', true);

                if ( $zoomMeetingID != '' ) {

                    $this->add_update_single_zoom_registrant($zoomMeetingID, $args);

                }

            }

        } elseif ( $WooCommerceEventsZoomMultiOption == 'multi' ) {

            // Multiple meetings
            $WooCommerceEventsZoomWebinarMulti = get_post_meta($ticket_meta['WooCommerceEventsProductID'][0], 'WooCommerceEventsZoomWebinarMulti', true);
            $WooCommerceEventsMultidayStatus = json_decode($ticket_meta['WooCommerceEventsMultidayStatus'][0], true);

            for ( $i = 1; $i <= count($WooCommerceEventsZoomWebinarMulti); $i++ ) {

                $zoomMeetingID = $WooCommerceEventsZoomWebinarMulti[$i - 1];

                if ( $zoomMeetingID != '' ) {

                    if ( empty($WooCommerceEventsMultidayStatus) || (!empty($WooCommerceEventsMultidayStatus) && $WooCommerceEventsMultidayStatus[(string)$i] != 'Canceled') ) {

                        $this->add_update_single_zoom_registrant($zoomMeetingID, $args);
                        
                    }

                }

            }

        } elseif ( $WooCommerceEventsZoomMultiOption == 'bookings' ) {

            // Bookable slot meeting
            $WooCommerceEventsBookingOptionsJSON = get_post_meta($WooCommerceEventsProductID, 'fooevents_bookings_options_serialized', true);
            $WooCommerceEventsBookingOptions = json_decode(!empty($WooCommerceEventsBookingOptionsJSON) ? $WooCommerceEventsBookingOptionsJSON : '{}', true);
            $WooCommerceEventsBookingSlotID = $ticket_meta['WooCommerceEventsBookingSlotID'][0];
            $WooCommerceEventsBookingDateID = $ticket_meta['WooCommerceEventsBookingDateID'][0];

            if ( isset($WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['zoom_id']) ) {

                $zoomMeetingID = $WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['add_date'][$WooCommerceEventsBookingDateID]['zoom_id'];

                $this->add_update_single_zoom_registrant($zoomMeetingID, $args);

            }

        }

    }

    /**
     * Generate text to display on the attendee's ticket
     * 
     * @param array $options
     * @param string $display
     * 
     * @return string
     */
    public function get_ticket_text($options = array(), $display = '') {

        $WooCommerceEventsProductID = $options['WooCommerceEventsProductID'];
        $registrant_email = !empty($options['registrant_email']) ? $options['registrant_email'] : '';
        $zoomTicketText = '';

        $WooCommerceEventsType = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsType', true);
        $WooCommerceEventsZoomMultiOption = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomMultiOption', true);

        if ( empty($WooCommerceEventsZoomMultiOption) || $WooCommerceEventsZoomMultiOption == 'single' || $WooCommerceEventsZoomMultiOption == 'bookings' ) {

            $zoomMeetingID = '';

            if ( $WooCommerceEventsZoomMultiOption == 'bookings' ) {

                // Bookable slot meeting
                $WooCommerceEventsBookingOptionsJSON = get_post_meta($WooCommerceEventsProductID, 'fooevents_bookings_options_serialized', true);
                $WooCommerceEventsBookingOptions = json_decode(!empty($WooCommerceEventsBookingOptionsJSON) ? $WooCommerceEventsBookingOptionsJSON : '{}', true);
                
                $WooCommerceEventsBookingSlotID = $options['slot_id'];
                $WooCommerceEventsBookingDateID = $options['date_id'];

                if ( isset($WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['zoom_id']) ) {

                    if ( isset($WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['add_date']) ) {

                        $zoomMeetingID = $WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID]['add_date'][$WooCommerceEventsBookingDateID]['zoom_id'];
                        
                    } else {

                        $zoomMeetingID = $WooCommerceEventsBookingOptions[$WooCommerceEventsBookingSlotID][$WooCommerceEventsBookingDateID . '_zoom_id'];

                    }

                }

            } else {

                // Single meeting
                $zoomMeetingID = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomWebinar', true);

            }

            if ( $zoomMeetingID != '' ) {

                $result = $this->do_fooevents_fetch_zoom_meeting($zoomMeetingID);

                if ( !empty($result['status']) && $result['status'] == 'success' ) {

                    $zoomMeeting = $result['data'];
                    $isMeeting = in_array((int)$zoomMeeting['type'], array(2, 3, 8));

                    if ( $display != 'admin' ) {
                        $zoomTicketText .= '<br/><br/>';

                        if ( $isMeeting ) {
                            $zoomTicketText .= '<strong>' . __('Zoom Meeting', 'woocommerce-events') . '</strong><br/>';
                        } else {
                            $zoomTicketText .= '<strong>' . __('Zoom Webinar', 'woocommerce-events') . '</strong><br/>';
                        }
                    }
                    
                    $zoomTicketText .= __('Topic', 'woocommerce-events') . ': ' . $zoomMeeting['topic'] . '<br/>';
                    $zoomTicketText .= (($zoomMeeting['type'] == 5 || $zoomMeeting['type'] == 2) ? __('Date', 'woocommerce-events') : __('Start date', 'woocommerce-events')) . ': ' . $zoomMeeting['start_date_display'] . '<br/>';
                    $zoomTicketText .= __('Start time', 'woocommerce-events') . ': ' . $zoomMeeting['start_time_display'] . '<br/>';
                    $zoomTicketText .= __('End time', 'woocommerce-events') . ': ' . $zoomMeeting['end_time_display'] . '<br/>';
                    $zoomTicketText .= __('Duration', 'woocommerce-events') . ': ' . $zoomMeeting['duration_display'] . '<br/>';

                    if ( $zoomMeeting['type'] != 5 && $zoomMeeting['type'] != 2 ) {

                        $zoomTicketText .= __('Recurrence', 'woocommerce-events') . ': ' . $zoomMeeting['recurrence']['type_display'] . '<br/>';

                    }

                    if ( $isMeeting ) {
                        $zoomTicketText .= __('Meeting ID', 'woocommerce-events') . ': ' . $this->format_zoom_id($zoomMeeting['id']) . '<br/>';
                        
                        if ( !empty($zoomMeeting['password']) ) {
                            $zoomTicketText .= __('Meeting password', 'woocommerce-events') . ': ' . $zoomMeeting['password'] . '<br/>';
                        }
                    } else {
                        $zoomTicketText .= __('Webinar ID', 'woocommerce-events') . ': ' . $this->format_zoom_id($zoomMeeting['id']) . '<br/>';

                        if ( !empty($zoomMeeting['password']) ) {
                            $zoomTicketText .= __('Webinar password', 'woocommerce-events') . ': ' . $zoomMeeting['password'] . '<br/>';
                        }
                    }
                    
                    $join_url = $zoomMeeting['join_url'];

                    if ( $registrant_email != '' ) {
                        
                        foreach ( $zoomMeeting['registrants']['registrants'] as $registrant ) {

                            if ( $registrant['email'] == $registrant_email ) {

                                $join_url = $registrant['join_url'];

                                break;

                            }

                        }

                    }

                    if ( $display == 'calendar' ) {

                        $zoomTicketText .= __('Join link', 'woocommerce-events') . ':<br/><a href="' . $join_url . '">' . $join_url . '</a><br/>';
                    
                    } elseif ( $display != 'admin' ) {

                        $zoomTicketText .= '<a href="' . $join_url . '">' . ($isMeeting ? __('Join this meeting', 'woocommerce-events') : __('Join this webinar', 'woocommerce-events')) . '</a><br/>';

                    }

                }

            }

        } elseif ( $WooCommerceEventsZoomMultiOption == 'multi' ) {

            // Multiple meetings
            $dayTerm = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsDayOverride', true);

            if(empty($dayTerm)) {

                $dayTerm = get_option('WooCommerceEventsDayOverride', true);

            }

            if(empty($dayTerm) || $dayTerm == 1) {

                $dayTerm = __('Day', 'woocommerce-events');

            }

            $WooCommerceEventsZoomWebinarMulti = get_post_meta($WooCommerceEventsProductID, 'WooCommerceEventsZoomWebinarMulti', true);

            if ( !empty($WooCommerceEventsZoomWebinarMulti) ) {

                if ( $display != 'admin' ) {
                    $zoomTicketText .= '<br/><br/>';
                    $zoomTicketText .= '<strong>' . __('Zoom Meetings and Webinars', 'woocommerce-events') . '</strong>';
                    $zoomTicketText .= '<br/>';
                }

                for ( $i = 1; $i <= count($WooCommerceEventsZoomWebinarMulti); $i++ ) {

                    $zoomMeetingID = $WooCommerceEventsZoomWebinarMulti[$i - 1];

                    if ( $zoomMeetingID != '' ) {

                        $result = $this->do_fooevents_fetch_zoom_meeting($zoomMeetingID);

                        if ( !empty($result['status']) && $result['status'] == 'success' ) {

                            $zoomMeeting = $result['data'];

                            $isMeeting = in_array((int)$zoomMeeting['type'], array(2, 3, 8));
        
                            if ( $i > 1 ) {
                                $zoomTicketText .= '<br/>';
                            }

                            $zoomTicketText .= '<strong>' . $dayTerm . ' ' . $i . ':</strong><br/>';
                            $zoomTicketText .= __('Topic', 'woocommerce-events') . ': ' . $zoomMeeting['topic'] . '<br/>';
                            
                            $zoomTicketText .= !empty($zoomMeeting['start_date_display']) ? (($zoomMeeting['type'] == 5 || $zoomMeeting['type'] == 2) ? __('Date', 'woocommerce-events') : __('Start date', 'woocommerce-events')) . ': ' . $zoomMeeting['start_date_display'] . '<br/>' : '';

                            $zoomTicketText .= !empty($zoomMeeting['start_time_display']) ? __('Start time', 'woocommerce-events') . ': ' . $zoomMeeting['start_time_display'] . '<br/>' : '';
                            $zoomTicketText .= !empty($zoomMeeting['end_time_display']) ? __('End time', 'woocommerce-events') . ': ' . $zoomMeeting['end_time_display'] . '<br/>' : '';
                            $zoomTicketText .= !empty($zoomMeeting['duration_display']) ? __('Duration', 'woocommerce-events') . ': ' . $zoomMeeting['duration_display'] . '<br/>' : '';

                            if ( $zoomMeeting['type'] != 5 && $zoomMeeting['type'] != 2 ) {

                                $zoomTicketText .= __('Recurrence', 'woocommerce-events') . ': ' . $zoomMeeting['recurrence']['type_display'] . '<br/>';

                            }
                            
                            if ( $isMeeting ) {
                                $zoomTicketText .= __('Meeting ID', 'woocommerce-events') . ': ' . $this->format_zoom_id($zoomMeeting['id']) . '<br/>';
                                
                                if ( !empty($zoomMeeting['password']) ) {
                                    $zoomTicketText .= __('Meeting password', 'woocommerce-events') . ': ' . $zoomMeeting['password'] . '<br/>';
                                }
                            } else {
                                $zoomTicketText .= __('Webinar ID', 'woocommerce-events') . ': ' . $this->format_zoom_id($zoomMeeting['id']) . '<br/>';
        
                                if ( !empty($zoomMeeting['password']) ) {
                                    $zoomTicketText .= __('Webinar password', 'woocommerce-events') . ': ' . $zoomMeeting['password'] . '<br/>';
                                }
                            }
                            
                            $join_url = $zoomMeeting['join_url'];

                            if ( $registrant_email != '' ) {
                                
                                foreach ( $zoomMeeting['registrants']['registrants'] as $registrant ) {

                                    if ( $registrant['email'] == $registrant_email ) {

                                        $join_url = $registrant['join_url'];

                                        break;

                                    }

                                }

                            }

                            if ( $display == 'calendar' ) {

                                $zoomTicketText .= __('Join link', 'woocommerce-events') . ':<br/><a href="' . $join_url . '">' . $join_url . '</a><br/>';
                            
                            } elseif ( $display != 'admin' ) {

                                $zoomTicketText .= '<a href="' . $join_url . '">' . ($isMeeting ? __('Join this meeting', 'woocommerce-events') : __('Join this webinar', 'woocommerce-events')) . '</a><br/>';

                            }

                        }
        
                    }

                }

            }

        }

        return $zoomTicketText;

    }

    /**
     * Generate text to display in the calendar event's description
     * 
     * @param array $options
     * 
     * @return string
     */
    public function get_calendar_text($options = array()) {

        $ticketText = $this->get_ticket_text($options, 'calendar');

        $ticketText = strip_tags(str_replace('<br/>', '\n', $ticketText));

        return $ticketText;

    }

    /**
     * Format Zoom meeting/webinar ID
     * 
     * @param int $zoomID
     * 
     * @return string
     */
    private function format_zoom_id($zoomID) {

        switch ( strlen($zoomID) ) {

            case 9:
            case 10:
            return substr($zoomID, 0, 3) . '-' . substr($zoomID, 3, 3) . '-' . substr($zoomID, 6);
            break;

            case 11:
            return substr($zoomID, 0, 3) . '-' . substr($zoomID, 3, 4) . '-' . substr($zoomID, 7);
            break;

        }

    }

    /**
     * Create a new Zoom meeting/webinar
     * 
     * @param array $options
     * 
     * @return string
     */
    public function create_zoom_meeting($options = array()) {

        $result = array('status' => 'error');

        if ( !empty($options) ) {

            $args = array_merge($options, array(
                'password' => $this->generate_zoom_passcode(),
                'settings' => array(
                    'approval_type' => 0,
                    'registration_type' => 1,
                    'close_registration' => true,
                    'registrants_email_notification' => true,
                    'waiting_room' => false
                )
            ));
            
            $endpoint = $options['type'];
            $type = $endpoint == 'meetings' ? 2 : 5;

            if ( (int)$options['type'] > 0 ) {

                $type = (int)$options['type'];

                $endpoint = in_array($type, array(2, 8)) ? 'meetings' : 'webinars';

            }
            
            $args['type'] = $type;

            $result = $this->fooevents_zoom_request('users/me/' . $endpoint, $args, '', 'POST');

            if ( $result['status'] === 'success' && !empty($result['data']['id']) ) {

                $result['data']['id'] .= '_' . $endpoint;

            }

        }

        return $result;

    }

    /**
     * Generate a random passcode.
     *
     * @since 1.12.17
     *
     * @return string
     */
    function generate_zoom_passcode() {

        $characters        = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ@-_*';
        $characters_length = strlen( $characters );
        $zoom_passcode     = '';

        for ( $i = 0; $i < 10; $i++ ) {

            $zoom_passcode .= $characters[ wp_rand( 0, $characters_length - 1 ) ];

        }

        return $zoom_passcode;

    }

    /**
     * Update an existing Zoom meeting/webinar
     * 
     * @param int $zoom_id
     * @param array $options
     * 
     * @return string
     */
    public function update_zoom_meeting($zoom_id, $options = array()) {

        $result = array('status' => 'error');

        if ( !empty($options) ) {

            $args = array_merge($options, array(
                'settings' => array(
                    'approval_type' => 0,
                    'registration_type' => 1
                )
            ));

            $zoomIDParts = explode("_", $zoom_id);
            $zoomID = $zoomIDParts[0];
            $endpoint = !empty($zoomIDParts[1]) ? $zoomIDParts[1] : 'webinars';

            $result = $this->fooevents_zoom_request($endpoint . '/' . $zoomID, $args, '', 'PATCH');

            if ( $result['status'] === 'success' ) {

                $result['data'] = array(
                    'id' => $zoom_id,
                    'topic' => $options['topic']
                );

            }

        }

        return $result;

    }

    /**
     * Delete an existing Zoom meeting/webinar
     * 
     * @param int $zoom_id
     */
    public function delete_zoom_meeting($zoom_id = '') {

        $result = array('status' => 'error');

        if ( !empty($zoom_id) ) {

            $zoomIDParts = explode("_", $zoom_id);
            $zoomID = $zoomIDParts[0];
            $endpoint = !empty($zoomIDParts[1]) ? $zoomIDParts[1] : 'webinars';

            $result = $this->fooevents_zoom_request($endpoint . '/' . $zoomID, array(), '', 'DELETE');

        }

        return $result;

    }

    /**
     * Add a user as an assistant (scheduler) to the main Zoom account holder
     * 
     * @param int $user_id
     */
    public function add_zoom_assistant($user_id = '') {

        $result = array('status' => 'error');

        if ( !empty($user_id) ) {

            $zoom_admin_user_id = get_option('globalWooCommerceEventsZoomAdminID');

            if ( empty($zoom_admin_user_id) ) {
                $result = $this->fooevents_zoom_request('users/me');

                if ( $result['status'] === 'success' ) {
                    $zoom_admin_user_id = $result['data']['id'];

                    update_option('globalWooCommerceEventsZoomAdminID', $zoom_admin_user_id);
                }
            }

            if ( !empty($zoom_admin_user_id) ) {
                $result = $this->fooevents_zoom_request('users/' . $user_id . '/assistants', array(
                    'assistants' => array(
                        array('id' => $zoom_admin_user_id)
                    )
                ), '', 'POST');
            }

        }

        return $result;

    }

    /**
     * Get the current or previous Zoom options for a single or multiday Zoom meeting integration from submitted product page save
     * 
     * @param int $post_id
     * @param boolean $previous
     * @param array $previous_post_meta
     * @param int $dayIndex
     * 
     * @return array
     */
    public function get_product_zoom_options($post_id, $previous = false, $previous_post_meta = array(), $dayIndex = -1) {

        $WooCommerceEventsType = $previous ? $previous_post_meta['WooCommerceEventsType'][0] : (isset($_POST['WooCommerceEventsType']) ? sanitize_text_field($_POST['WooCommerceEventsType']) : 'single');
        $WooCommerceEventsZoomMultiOption = $previous ? $previous_post_meta['WooCommerceEventsZoomMultiOption'][0] : sanitize_text_field($_POST['WooCommerceEventsZoomMultiOption']);

        if ( $previous && empty($previous_post_meta['WooCommerceEventsZoomType']) ) {

            return array();

        }
        
        $WooCommerceEventsZoomHost = $previous ? $previous_post_meta['WooCommerceEventsZoomHost'][0] : (isset($_POST['WooCommerceEventsZoomHost']) ? sanitize_text_field($_POST['WooCommerceEventsZoomHost']) : 'me');
        
        $topic = html_entity_decode($previous ? apply_filters('the_title', sanitize_text_field(stripslashes($_POST['WooCommerceEventsZoomTopic'])), $post_id) : apply_filters('the_title', sanitize_text_field(stripslashes($_POST['post_title'])), $post_id));

        if ( $dayIndex > -1 ) {

            if ( $WooCommerceEventsType == 'sequential' ) {

                $tempStartDate = $previous ? $previous_post_meta['WooCommerceEventsDate'][0] : (isset($_POST['WooCommerceEventsDate']) ? sanitize_text_field($_POST['WooCommerceEventsDate']) : '');

                $startDateTimestamp = strtotime($this->convert_month_to_english($tempStartDate) . ' + ' . $dayIndex . ' days');
                $startDate = date(get_option('date_format'), $startDateTimestamp);

            } elseif ( $WooCommerceEventsType == 'select' ) {

                $startDate = $previous ? unserialize($previous_post_meta['WooCommerceEventsSelectDate'][0])[$dayIndex] : (isset($_POST['WooCommerceEventsSelectDate']) ? sanitize_text_field($_POST['WooCommerceEventsSelectDate'][$dayIndex]) : '');

            }

        } else {

            $startDate = $previous ? $previous_post_meta['WooCommerceEventsDate'][0] : (isset($_POST['WooCommerceEventsDate']) ? sanitize_text_field($_POST['WooCommerceEventsDate']) : '');

        }
        
        $startTimeHour = $previous ? $previous_post_meta['WooCommerceEventsHour'][0] : (isset($_POST['WooCommerceEventsHour']) ? sanitize_text_field($_POST['WooCommerceEventsHour']) : '00');
        $startTimeMinutes = $previous ? $previous_post_meta['WooCommerceEventsMinutes'][0] : (isset($_POST['WooCommerceEventsMinutes']) ? sanitize_text_field($_POST['WooCommerceEventsMinutes']) : '00');
        $startTimePeriod = $previous ? $previous_post_meta['WooCommerceEventsPeriod'][0] : (isset($_POST['WooCommerceEventsPeriod']) ? sanitize_text_field($_POST['WooCommerceEventsPeriod']) : '');
        $endTimeHour = $previous ? $previous_post_meta['WooCommerceEventsHourEnd'][0] : (isset($_POST['WooCommerceEventsHourEnd']) ? sanitize_text_field($_POST['WooCommerceEventsHourEnd']) : '01');
        $endTimeMinutes = $previous ? $previous_post_meta['WooCommerceEventsMinutesEnd'][0] : (isset($_POST['WooCommerceEventsMinutesEnd']) ? sanitize_text_field($_POST['WooCommerceEventsMinutesEnd']) : '00');
        $endTimePeriod = $previous ? $previous_post_meta['WooCommerceEventsEndPeriod'][0] : (isset($_POST['WooCommerceEventsEndPeriod']) ? sanitize_text_field($_POST['WooCommerceEventsEndPeriod']) : '');

        $duration = 60;

        if ( ($startTimePeriod == "" && $endTimePeriod == "") || $startTimePeriod == $endTimePeriod ) {

            $duration = (((int)$endTimeHour * 60) + (int)$endTimeMinutes) - (((int)$startTimeHour * 60) + (int)$startTimeMinutes);

        } else {

            $duration = ((((int)$endTimeHour + ($endTimePeriod == "p.m." && (int)$endTimeHour < 12 ? 12 : 0)) * 60) + (int)$endTimeMinutes) - (((int)$startTimeHour * 60) + (int)$startTimeMinutes);
        }
        
        $endpoint = $previous ? $previous_post_meta['WooCommerceEventsZoomType'][0] : (isset($_POST['WooCommerceEventsZoomType']) ? sanitize_text_field($_POST['WooCommerceEventsZoomType']) : 'meetings');
        $type = $endpoint == 'meetings' ? ($WooCommerceEventsType == 'sequential' && $WooCommerceEventsZoomMultiOption == 'single' ? 8 : 2) : ($WooCommerceEventsType == 'sequential' && $WooCommerceEventsZoomMultiOption == 'single' ? 9 : 5);

        $start_time_display = $startDate . ' ' . $startTimeHour . ':' . $startTimeMinutes . ($startTimePeriod != '' ? ' ' . $startTimePeriod : '');
        $start_timestamp = strtotime($this->convert_month_to_english($start_time_display));

        $WooCommerceEventsTimeZone = $previous ? $previous_post_meta['WooCommerceEventsTimeZone'][0] : (isset($_POST['WooCommerceEventsTimeZone']) ? sanitize_text_field($_POST['WooCommerceEventsTimeZone']) : get_option('timezone_string'));
        
        $zoom_options = array(
            'topic' => $topic . ($dayIndex > -1 ? ' - ' . $start_time_display : ''),
            'timezone' => $WooCommerceEventsTimeZone,
            'schedule_for' => $WooCommerceEventsZoomHost,
            'duration' => $duration,
            'type' => $type,
            'start_time' => date('Y-m-d\TH:i:s', $start_timestamp),
        );

        if ( $WooCommerceEventsType == 'sequential' && $WooCommerceEventsZoomMultiOption == 'single' ) {

            $numDays = $previous ? (int)$previous_post_meta['WooCommerceEventsNumDays'][0] : (isset($_POST['WooCommerceEventsNumDays']) ? (int)sanitize_text_field($_POST['WooCommerceEventsNumDays']) : 1);

            $zoom_options['recurrence'] = array(
                'type' => 1,
                'repeat_interval' => 1,
                'end_times' => $numDays,
            );
        }

        return $zoom_options;

    }

    /**
     * Array of month names for translation to English
     * 
     * @param string $event_date
     * @return string
     */
    private function convert_month_to_english($event_date) {
        
        $months = array(
            // French
            'janvier' => 'January',
            'février' => 'February',
            'mars' => 'March',
            'avril' => 'April',
            'mai' => 'May',
            'juin' => 'June',
            'juillet' => 'July',
            'aout' => 'August',
            'août' => 'August',
            'septembre' => 'September',
            'octobre' => 'October',

            // German
            'Januar' => 'January',
            'Februar' => 'February',
            'März' => 'March',
            'Mai' => 'May',
            'Juni' => 'June',
            'Juli' => 'July',
            'Oktober' => 'October',
            'Dezember' => 'December',

            // Spanish
            'enero' => 'January',
            'febrero' => 'February',
            'marzo' => 'March',
            'abril' => 'April',
            'mayo' => 'May',
            'junio' => 'June',
            'julio' => 'July',
            'agosto' => 'August',
            'septiembre' => 'September',
            'setiembre' => 'September',
            'octubre' => 'October',
            'noviembre' => 'November',
            'diciembre' => 'December',
            'novembre' => 'November',
            'décembre' => 'December',

            // Catalan - Spain
            "gener" => "January",
            "febrer" => "February",
            "març" => "March",
            "abril" => "April",
            "maig" => "May",
            "juny" => "June",
            "juliol" => "July",
            "agost" => "August",
            "setembre" => "September",
            "octubre" => "October",
            "novembre" => "November",
            "desembre" => "December",

            // Dutch
            'januari' => 'January',
            'februari' => 'February',
            'maart' => 'March',
            'april' => 'April',
            'mei' => 'May',
            'juni' => 'June',
            'juli' => 'July',
            'augustus' => 'August',
            'september' => 'September',
            'oktober' => 'October',
            'november' => 'November',
            'december' => 'December',

            // Italian
            'Gennaio' => 'January',
            'Febbraio' => 'February',
            'Marzo' => 'March',
            'Aprile' => 'April',
            'Maggio' => 'May',
            'Giugno' => 'June',
            'Luglio' => 'July',
            'Agosto' => 'August',
            'Settembre' => 'September',
            'Ottobre' => 'October',
            'Novembre' => 'November',
            'Dicembre' => 'December',

            // Polish
            'Styczeń' => 'January',
            'Luty' => 'February',
            'Marzec' => 'March',
            'Kwiecień' => 'April',
            'Maj' => 'May',
            'Czerwiec' => 'June',
            'Lipiec' => 'July',
            'Sierpień' => 'August',
            'Wrzesień' => 'September',
            'Październik' => 'October',
            'Listopad' => 'November',
            'Grudzień' => 'December',

            // Afrikaans
            'Januarie' => 'January',
            'Februarie' => 'February',
            'Maart' => 'March',
            'Mei' => 'May',
            'Junie' => 'June',
            'Julie' => 'July',
            'Augustus' => 'August',
            'Oktober' => 'October',
            'Desember' => 'December',

            // Turkish
            'Ocak' => 'January',
            'Şubat' => 'February',
            'Mart' => 'March',
            'Nisan' => 'April',
            'Mayıs' => 'May',
            'Haziran' => 'June',
            'Temmuz' => 'July',
            'Ağustos' => 'August',
            'Eylül' => 'September',
            'Ekim' => 'October',
            'Kasım' => 'November',
            'Aralık' => 'December',

            // Portuguese
            'janeiro' => 'January',
            'fevereiro' => 'February',
            'março' => 'March',
            'abril' => 'April',
            'maio' => 'May',
            'junho' => 'June',
            'julho' => 'July',
            'agosto' => 'August',
            'setembro' => 'September',
            'outubro' => 'October',
            'novembro' => 'November',
            'dezembro' => 'December',

            // Swedish
            'Januari' => 'January',
            'Februari' => 'February',
            'Mars' => 'March',
            'April' => 'April',
            'Maj' => 'May',
            'Juni' => 'June',
            'Juli' => 'July',
            'Augusti' => 'August',
            'September' => 'September',
            'Oktober' => 'October',
            'November' => 'November',
            'December' => 'December',

            // Czech
            'leden' => 'January',
            'únor' => 'February',
            'březen' => 'March',
            'duben' => 'April',
            'květen' => 'May',
            'červen' => 'June',
            'červenec' => 'July',
            'srpen' => 'August',
            'září' => 'September',
            'říjen' => 'October',
            'listopad' => 'November',
            'prosinec' => 'December',

            // Norwegian
            'januar' => 'January',
            'februar' => 'February',
            'mars' => 'March',
            'april' => 'April',
            'mai' => 'May',
            'juni' => 'June',
            'juli' => 'July',
            'august' => 'August',
            'september' => 'September',
            'oktober' => 'October',
            'november' => 'November',
            'desember' => 'December',

            // Danish
            'januar' => 'January',
            'februar' => 'February',
            'marts' => 'March',
            'april' => 'April',
            'maj' => 'May',
            'juni' => 'June',
            'juli' => 'July',
            'august' => 'August',
            'september' => 'September',
            'oktober' => 'October',
            'november' => 'November',
            'december' => 'December',

            // Finnish
            'tammikuu' => 'January',
            'helmikuu' => 'February',
            'maaliskuu' => 'March',
            'huhtikuu' => 'April',
            'toukokuu' => 'May',
            'kesäkuu' => 'June',
            'heinäkuu' => 'July',
            'elokuu' => 'August',
            'syyskuu' => 'September',
            'lokakuu' => 'October',
            'marraskuu' => 'November',
            'joulukuu' => 'December',

            // Russian
            'Январь' => 'January', 
            'Февраль' => 'February', 
            'Март' => 'March', 
            'Апрель' => 'April', 
            'Май' => 'May', 
            'Июнь' => 'June', 
            'Июль' => 'July', 
            'Август' => 'August', 
            'Сентябрь' => 'September', 
            'Октябрь' => 'October', 
            'Ноябрь' => 'November', 
            'Декабрь' => 'December',

            // Icelandic
            'Janúar' => 'January', 
            'Febrúar' => 'February', 
            'Mars' => 'March', 
            'Apríl' => 'April', 
            'Maí' => 'May', 
            'Júní' => 'June', 
            'Júlí' => 'July', 
            'Ágúst' => 'August', 
            'September' => 'September', 
            'Oktober' => 'October', 
            'Nóvember' => 'November', 
            'Desember' => 'December',

            // Latvian
            'janvāris' => 'January',
            'februāris' => 'February',
            'marts' => 'March',
            'aprīlis' => 'April',
            'maijs' => 'May',
            'jūnijs' => 'June',
            'jūlijs' => 'July',
            'augusts' => 'August',
            'septembris' => 'September',
            'oktobris' => 'October',
            'novembris' => 'November',
            'decembris' => 'December',

            // Lithuanian
            'Sausis' => 'January',
            'Vasaris' => 'February',
            'Kovas' => 'March',
            'Balandis' => 'April',
            'Gegužė' => 'May',
            'Birželis' => 'June',
            'Liepa' => 'July',
            'Rugpjūtis' => 'August',
            'Rugsėjis' => 'September',
            'Spalis' => 'October',
            'Lapkritis' => 'November',
            'Gruodis' => ' December',

            // Greek
            'Ιανουάριος' => 'January',
            'Φεβρουάριος' => 'February',
            'Μάρτιος' => 'March',
            'Απρίλιος' => 'April',
            'Μάιος' => 'May',
            'Ιούνιος' => 'June',
            'Ιούλιος' => 'July',
            'Αύγουστος' => 'August',
            'Σεπτέμβριος' => 'September',
            'Οκτώβριος' => 'October',
            'Νοέμβριος' => 'November',
            'Δεκέμβριος' => 'December',

            // Slovak - Slovakia
            "január" => "January",
            "február" => "February",
            "marec" => "March",
            "apríl" => "April",
            "máj" => "May",
            "jún" => "June",
            "júl" => "July",
            "august" => "August",
            "september" => "September",
            "október" => "October",
            "november" => "November",
            "december" => "December",

            // Slovenian - Slovenia
            "januar" => "January",
            "februar" => "February",
            "marec" => "March",
            "april" => "April",
            "maj" => "May",
            "junij" => "June",
            "julij" => "July",
            "avgust" => "August",
            "september" => "September",
            "oktober" => "October",
            "november" => "November",
            "december" => "December",
        );
        
        $pattern = array_keys($months);
        $replacement = array_values($months);

        foreach ($pattern as $key => $value) {
            
            $pattern[$key] = '/\b'.$value.'\b/i';
            
        }
        
        $replaced_event_date = preg_replace($pattern, $replacement, $event_date);

        $replaced_event_date = str_replace(' de ', ' ', $replaced_event_date);

        return $replaced_event_date;
        
    }

}